import graphene
import graphene_django_optimizer as gql_optimizer
from core.gql_queries import ValidationMessageGQLType
from core.schema import (
    OrderedDjangoFilterConnectionField,
    signal_mutation_module_before_mutating,
)
from core.utils import append_validity_filter
from core.services import wait_for_mutation
from django.db.models import Q

from contract.apps import ContractConfig
from contract.gql.gql_mutations.contract_details_mutations import (
    CreateContractDetailByPolicyHolderInsureeMutation,
    CreateContractDetailsMutation,
    DeleteContractDetailsMutation,
    UpdateContractDetailsMutation,
)
from contract.gql.gql_mutations.contract_mutations import (
    AmendContractMutation,
    ApproveContractBulkMutation,
    ApproveContractMutation,
    ContractCreateInvoiceBulkMutation,
    CounterContractBulkMutation,
    CounterContractMutation,
    CreateContractMutation,
    DeleteContractMutation,
    RenewContractMutation,
    SubmitContractMutation,
    UpdateContractMutation,
)
from contract.gql.gql_types import (
    ContractContributionPlanDetailsGQLType,
    ContractDetailsGQLType,
    ContractGQLType,
)
from contract.models import (
    Contract,
    ContractContributionPlanDetails,
    ContractDetails,
    ContractMutation,
)
from contract.utils import filter_amount_contract

from .services import check_unique_code


class Query(graphene.ObjectType):

    contract = OrderedDjangoFilterConnectionField(
        ContractGQLType,
        client_mutation_id=graphene.String(),
        insuree=graphene.UUID(),
        orderBy=graphene.List(of_type=graphene.String),
        dateValidFrom__Gte=graphene.DateTime(),
        dateValidTo__Lte=graphene.DateTime(),
        amount_from=graphene.Decimal(),
        amount_to=graphene.Decimal(),
        applyDefaultValidityFilter=graphene.Boolean(),
    )

    contract_details = OrderedDjangoFilterConnectionField(
        ContractDetailsGQLType,
        client_mutation_id=graphene.String(),
        orderBy=graphene.List(of_type=graphene.String),
    )

    contract_contribution_plan_details = OrderedDjangoFilterConnectionField(
        ContractContributionPlanDetailsGQLType,
        insuree=graphene.UUID(),
        contributionPlanBundle=graphene.UUID(),
        orderBy=graphene.List(of_type=graphene.String),
    )

    validate_contract_code = graphene.Field(
        ValidationMessageGQLType,
        contract_code=graphene.String(required=True),
        description="Check that the specified contract code is unique.",
    )

    def resolve_validate_contract_code(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(
                ContractConfig.gql_query_contract_policyholder_portal_perms
            ):
                raise PermissionError("Unauthorized")
        errors = check_unique_code(code=kwargs["contract_code"])
        if errors:
            return ValidationMessageGQLType(False)
        else:
            return ValidationMessageGQLType(True)

    def resolve_contract(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(
                ContractConfig.gql_query_contract_policyholder_portal_perms
            ):
                raise PermissionError("Unauthorized")

        filters = append_validity_filter(**kwargs)
        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(
                Q(mutations__mutation__client_mutation_id=client_mutation_id)
            )

        insuree = kwargs.get("insuree", None)
        if insuree:
            filters.append(Q(contractdetails__insuree__uuid=insuree))

        # amount filters
        amount_from = kwargs.get("amount_from", None)
        amount_to = kwargs.get("amount_to", None)
        if amount_from or amount_to:
            filters.append(filter_amount_contract(**kwargs))
        return gql_optimizer.query(Contract.objects.filter(*filters).all(), info)

    def resolve_contract_details(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(
                ContractConfig.gql_query_contract_policyholder_portal_perms
            ):
                raise PermissionError("Unauthorized")

        filters = []
        client_mutation_id = kwargs.get("client_mutation_id", None)
        if client_mutation_id:
            wait_for_mutation(client_mutation_id)
            filters.append(
                Q(mutations__mutation__client_mutation_id=client_mutation_id)
            )

        return gql_optimizer.query(ContractDetails.objects.filter(*filters).all(), info)

    def resolve_contract_contribution_plan_details(self, info, **kwargs):
        if not info.context.user.has_perms(ContractConfig.gql_query_contract_perms):
            if not info.context.user.has_perms(
                ContractConfig.gql_query_contract_policyholder_portal_perms
            ):
                raise PermissionError("Unauthorized")

        query = ContractContributionPlanDetails.objects.all()

        insuree = kwargs.get("insuree", None)
        contribution_plan_bundle = kwargs.get("contributionPlanBundle", None)

        if insuree:
            query = query.filter(contract_details__insuree__uuid=insuree)

        if contribution_plan_bundle:
            query = query.filter(
                contract_details__contribution_plan_bundle__id=contribution_plan_bundle
            )

        return gql_optimizer.query(query.all(), info)


class Mutation(graphene.ObjectType):
    create_contract = CreateContractMutation.Field()
    update_contract = UpdateContractMutation.Field()
    delete_contract = DeleteContractMutation.Field()
    submit_contract = SubmitContractMutation.Field()
    approve_contract = ApproveContractMutation.Field()
    approve_bulk_contract = ApproveContractBulkMutation.Field()
    counter_contract = CounterContractMutation.Field()
    counter_bulk_contract = CounterContractBulkMutation.Field()
    amend_contract = AmendContractMutation.Field()
    renew_contract = RenewContractMutation.Field()
    create_contract_invoice_bulk = ContractCreateInvoiceBulkMutation.Field()

    create_contract_details = CreateContractDetailsMutation.Field()
    update_contract_details = UpdateContractDetailsMutation.Field()
    delete_contract_details = DeleteContractDetailsMutation.Field()
    create_contract_details_by_ph_insuree = (
        CreateContractDetailByPolicyHolderInsureeMutation.Field()
    )


def on_contract_mutation(sender, **kwargs):
    uuids = kwargs["data"].get("uuids", [])
    if not uuids:
        uuid = kwargs["data"].get("uuid", None)
        uuids = [uuid] if uuid else []
    if not uuids:
        return []
    impacted_contracts = Contract.objects.filter(id__in=uuids).all()
    for contract in impacted_contracts:
        ContractMutation.objects.update_or_create(
            contract=contract, mutation_id=kwargs["mutation_log_id"]
        )
    return []


signal_mutation_module_before_mutating["contract"].connect(on_contract_mutation)



import logging

from django.db import migrations

from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)


def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 152101, apps)  # contract search
    insert_role_right_for_system(64, 152102, apps)  # contract create
    insert_role_right_for_system(64, 152103, apps)  # contract update
    insert_role_right_for_system(64, 152104, apps)  # contract delete
    insert_role_right_for_system(64, 152106, apps)  # contract renew
    insert_role_right_for_system(64, 152107, apps)  # contract submit
    insert_role_right_for_system(64, 152108, apps)  # contract approve


class Migration(migrations.Migration):
    dependencies = [
        ('contract', '0016_auto_20210208_1508')
    ]

    operations = [
        migrations.RunPython(add_rights)
    ]



from core.gql.gql_mutations import ObjectNotExistException
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ValidationError
from django.forms.models import model_to_dict
from django.utils.translation import gettext as _

from contract.apps import ContractConfig
from contract.models import Contract, ContractMutation
from contract.services import (
    Contract as ContractService,
    ContractDetails as ContractDetailsService,
    ContractToInvoiceService,
    _output_result_success,
    _output_exception
)


class ContractCreateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_create_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.create_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(
                user, client_mutation_id=client_mutation_id, contract=contract
            )
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def create_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.create(contract=contract)
        return output_data


class ContractUpdateMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_update_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.update_contract(user=user, contract=data)
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def update_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.update(contract=contract)
        return output_data


class ContractDeleteMutationMixin:
    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _object_not_exist_exception(cls, obj_uuid):
        raise ObjectNotExistException(cls._model, obj_uuid)

    @classmethod
    def _validate_mutation(cls, user, **data):
        cls._validate_user(user)

    @classmethod
    def _validate_user(cls, user):
        if type(user) is AnonymousUser or not user.id:
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, uuid):
        output = cls.delete_contract(user=user, contract={"id": uuid})
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def delete_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.delete(contract=contract)
        return output_data


class ContractSubmitMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_submit_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.submit_contract(user=user, contract=data)
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def submit_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.submit(contract=contract)
        return output_data


class ContractApproveMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(
                ContractConfig.gql_mutation_approve_ask_for_change_contract_perms
            )
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.approve_contract(user=user, contract=data)
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def approve_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.approve(contract=contract)
        return output_data


class ContractCounterMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(
                ContractConfig.gql_mutation_approve_ask_for_change_contract_perms
            )
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.counter_contract(user=user, contract=data)
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def counter_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.counter(contract=contract)
        return output_data


class ContractAmendMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_amend_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.amend_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(
                user, client_mutation_id=client_mutation_id, contract=contract
            )
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def amend_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.amend(contract=contract)
        return output_data


class ContractRenewMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_renew_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.renew_contract(user=user, contract=data)
        if output["success"]:
            contract = Contract.objects.get(id=output["data"]["id"])
            ContractMutation.object_mutated(
                user, client_mutation_id=client_mutation_id, contract=contract
            )
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def renew_contract(cls, user, contract):
        contract_service = ContractService(user=user)
        output_data = contract_service.renew(contract=contract)
        return output_data


class ContractDetailsFromPHInsureeMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_mutation_update_contract_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.create_cd_from_ph_insuree(user=user, data=data)
        return (
            None
            if output["success"]
            else f"Error! - {output['message']}: {output['detail']}"
        )

    @classmethod
    def create_cd_from_ph_insuree(cls, user, data):
        contract_details_service = ContractDetailsService(user=user)
        try:
            if not user.has_perms(
                ContractConfig.gql_mutation_update_contract_perms
            ):
                raise PermissionError("Unauthorized")
            contract = Contract.get(id=f'{data["contract_id"]}')
            if contract.state not in [
                    Contract.STATE_DRAFT,
                    Contract.STATE_REQUEST_FOR_INFORMATION,
                    Contract.STATE_COUNTER,
            ]:
                raise Exception(
                    _(
                        "You cannot update contract by adding insuree - contract not in updatable state!"
                    )
                )
            contract_details = contract_details_service.get_details_from_ph_insuree(
                contract,
                ph_insuree_id=data["policy_holder_insuree_id"]
            )
            dict_representation = []
            for cd in contract_details:
                uuid_string = f"{cd.id}"
                record = model_to_dict(cd)
                record["id"], record["uuid"] = (
                    uuid_string,
                    uuid_string,
                )
                dict_representation.append(record)
                return _output_result_success(dict_representation)
            else:
                raise Exception(
                    _(
                        "You cannot insuree - is deleted or not enough data to create contract!"
                    )
                )
        except Exception as exc:
            return _output_exception(
                model_name="ContractDetails",
                method="PHInsureToCDetatils",
                exception=exc,
            )


class ContractCreateInvoiceMutationMixin:

    @property
    def _model(self):
        raise NotImplementedError()

    @classmethod
    def _validate_mutation(cls, user, **data):
        if (
            type(user) is AnonymousUser
            or not user.id
            or not user.has_perms(ContractConfig.gql_invoice_create_perms)
        ):
            raise ValidationError("mutation.authentication_required")

    @classmethod
    def _mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        output = cls.create_contract_invoice(user=user, data=data)
        if output["success"]:
            return None
        else:
            return f"Error! - {output['message']}: {output['detail']}"

    @classmethod
    def create_contract_invoice(cls, user, data):
        queryset = Contract.objects.filter(id=data["id"])
        if queryset.count() == 1:
            contract = queryset.first()
            contract_to_invoice_service = ContractToInvoiceService(user=user)
            output_data = contract_to_invoice_service.create_invoice(
                instance=contract, convert_to="InvoiceLine", user=user
            )
            return output_data



import logging

from django.db import migrations

from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)


def add_contract(apps, schema_editor):
    insert_role_right_for_system(256, 152108, apps)


class Migration(migrations.Migration):
    dependencies = [
        ('contract', '0017_contract_roles_for_admin')
    ]

    operations = [
        migrations.RunPython(add_contract, None)
    ]



# Generated by Django 4.2.16 on 2024-10-07 12:10
import logging

from django.db import migrations

from core.utils import insert_role_right_for_system, remove_role_right_for_system

logger = logging.getLogger(__name__)

def add_rights(apps, schema_editor):
    insert_role_right_for_system(64, 101500, apps)
    insert_role_right_for_system(64, 154901, apps)

def remove_rights(apps, schema_editor):
    remove_role_right_for_system(64, 101500, apps)
    remove_role_right_for_system(64, 154901, apps)

class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0021_alter_contract_date_created_and_more'),
    ]

    operations = [
        migrations.RunPython(add_rights, reverse_code=remove_rights),
    ]



# Generated by Django 3.0.3 on 2020-12-30 10:52

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('policyholder', '0004_auto_20201214_1302'),
        ('contract', '0007_auto_20201228_1310'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contract',
            name='policy_holder',
            field=models.ForeignKey(blank=True, db_column='PolicyHolderUUID', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='policyholder.PolicyHolder'),
        ),
    ]



CONTRACT_STATE = {
    "ContractState": [
        {
            "value": "1",
            "label": {"fr": "Demande d'information", "en": "Request for information"},
        },
        {"value": "2", "label": {"fr": "Brouillon", "en": "Draft"}},
        {"value": "3", "label": {"fr": "Offre", "en": "offer"}},
        {"value": "4", "label": {"fr": "En negociation", "en": "Negotiable"}},
        {"value": "5", "label": {"fr": "Apprové", "en": "executable"}},
        {"value": "6", "label": {"fr": "addendum", "en": "addendum"}},
        {"value": "7", "label": {"fr": "En cours", "en": "effective"}},
        {"value": "8", "label": {"fr": "Appliqué", "en": "executed"}},
        {"value": "9", "label": {"fr": "Suspendu", "en": "Disputed"}},
        {"value": "10", "label": {"fr": "Terminé", "en": "terminated"}},
        {"value": "11", "label": {"fr": "révision demandé", "en": "counter"}},
    ]
}


def get_message_approved_contract(
    code, name, contact_name, due_amount, payment_reference, language="en"
):
    message_payment_notification = {
        "payment_notification": {
            "en": f"""
                 Dear {contact_name}

                 The contract {code} - {name} was approved.
                 Please proceed to the payment of {due_amount} with the reference {payment_reference}.

                 Best regards,
                 """,
            "fr": f"""
                 Monsieur, Madame {contact_name}

                 le contract {code} - {name} à été apprové.
                 Veuillez faire un paiement de < CONTRACT - DUEAMOUNT > avec la référence {payment_reference}.

                 Meilleurs Salutations
                 """,
        }
    }
    return message_payment_notification["payment_notification"][language]


def get_message_counter_contract(code, name, contact_name, language="en"):
    message_payment_notification = {
        "payment_notification": {
            "en": f"""
                 Dear {contact_name}

                 The contract {code} - {name} was counter.
                 Please proceed recheck the information and correct the issues,
                 in case of questions please check contact us.

                 Best regards,
                 """,
            "fr": f"""
                 Monsieur, Madame {contact_name}

                 le contract {code} - {name} à été contré.
                 Veuillez verifier les information saisie, en cas de question veuillez nous contacter.

                 Meilleurs Salutations
                 """,
        }
    }
    return message_payment_notification["payment_notification"][language]



from core.gql.gql_mutations import (
    DeleteInputType,
    mutation_on_uuids_from_filter_business_model,
)
from core.gql.gql_mutations.base_mutation import BaseDeleteMutation, BaseMutation
from kombu.exceptions import OperationalError

from contract.exceptions import CeleryWorkerError
from contract.gql.gql_mutations.input_types import (
    ContractAmendInputType,
    ContractApproveBulkInputType,
    ContractApproveInputType,
    ContractCounterBulkInputType,
    ContractCounterInputType,
    ContractCreateInputType,
    ContractCreateInvoiceBulkInputType,
    ContractRenewInputType,
    ContractSubmitInputType,
    ContractUpdateInputType,
)
from contract.gql.gql_types import ContractGQLType
from contract.models import Contract
from contract.tasks import (
    approve_contracts,
    counter_contracts,
    create_invoice_from_contracts,
)

from .mutations import (
    ContractAmendMutationMixin,
    ContractApproveMutationMixin,
    ContractCounterMutationMixin,
    ContractCreateInvoiceMutationMixin,
    ContractCreateMutationMixin,
    ContractDeleteMutationMixin,
    ContractRenewMutationMixin,
    ContractSubmitMutationMixin,
    ContractUpdateMutationMixin,
)


class CreateContractMutation(ContractCreateMutationMixin, BaseMutation):
    _mutation_class = "CreateContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractCreateInputType):
        pass


class UpdateContractMutation(ContractUpdateMutationMixin, BaseMutation):
    _mutation_class = "UpdateContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractUpdateInputType):
        pass


class DeleteContractMutation(ContractDeleteMutationMixin, BaseDeleteMutation):
    _mutation_class = "DeleteContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(DeleteInputType):
        pass


class SubmitContractMutation(ContractSubmitMutationMixin, BaseMutation):
    _mutation_class = "SubmitContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractSubmitInputType):
        pass


class ApproveContractMutation(ContractApproveMutationMixin, BaseMutation):
    _mutation_class = "ApproveContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractApproveInputType):
        pass


class ApproveContractBulkMutation(ContractApproveMutationMixin, BaseMutation):
    _mutation_class = "ApproveContractBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(
        Contract, ContractGQLType, "extended_filters", {}
    )
    def async_mutate(cls, user, **data):
        error_message = None
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        if "contract_uuids" in data or "uuids" in data:
            error_message = cls.approve_contracts(user=user, contracts=data)
        return error_message

    def _check_celery_status(cls):
        try:
            from openIMIS.celery import app as celery_app

            connection = celery_app.broker_connection().ensure_connection(max_retries=3)
            if not connection:
                raise CeleryWorkerError(
                    "Celery worker not found. Please, contact your system administrator."
                )
        except (IOError, OperationalError) as e:
            raise CeleryWorkerError(
                f"Celery connection has failed. Error: {e} \n Please, contact your system administrator."
            )

    @classmethod
    def approve_contracts(cls, user, contracts):

        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            return approve_contracts(user_id=f"{user.id}", contracts=contracts["uuids"])
        else:
            if "contract_uuids" in contracts:
                return approve_contracts(
                    user_id=f"{user.id}", contracts=contracts["contract_uuids"]
                )

    class Input(ContractApproveBulkInputType):
        pass


class CounterContractMutation(ContractCounterMutationMixin, BaseMutation):
    _mutation_class = "CounterContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractCounterInputType):
        pass


class ContractCreateInvoiceBulkMutation(
    ContractCreateInvoiceMutationMixin, BaseMutation
):
    _mutation_class = "ContractCreateInvoiceBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(
        Contract, ContractGQLType, "extended_filters", {}
    )
    def async_mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        if "contract_uuids" in data or "uuids" in data:
            cls.create_contract_invoice(user=user, contracts=data)
        return None

    @classmethod
    def create_contract_invoice(cls, user, contracts):
        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            return create_invoice_from_contracts(
                user_id=f"{user.id}", contracts=contracts["uuids"]
            )
        else:
            if "contract_uuids" in contracts:
                return create_invoice_from_contracts(
                    user_id=f"{user.id}", contracts=contracts["contract_uuids"]
                )

    class Input(ContractCreateInvoiceBulkInputType):
        pass


class CounterContractBulkMutation(ContractCounterMutationMixin, BaseMutation):
    _mutation_class = "CounterContractBulkMutation"
    _mutation_module = "contract"
    _model = Contract

    @classmethod
    @mutation_on_uuids_from_filter_business_model(
        Contract, ContractGQLType, "extended_filters", {}
    )
    def async_mutate(cls, user, **data):
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        if "contract_uuids" in data or "uuids" in data:
            cls.counter_contracts(user=user, contracts=data)
        return None

    @classmethod
    def counter_contracts(cls, user, contracts):
        if "uuids" in contracts:
            contracts["uuids"] = list(contracts["uuids"].values_list("id", flat=True))
            return counter_contracts(user_id=f"{user.id}", contracts=contracts["uuids"])
        else:
            if "contract_uuids" in contracts:
                return counter_contracts(
                    user_id=f"{user.id}", contracts=contracts["contract_uuids"]
                )

    class Input(ContractCounterBulkInputType):
        pass


class AmendContractMutation(ContractAmendMutationMixin, BaseMutation):
    _mutation_class = "AmendContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractAmendInputType):
        pass


class RenewContractMutation(ContractRenewMutationMixin, BaseMutation):
    _mutation_class = "RenewContractMutation"
    _mutation_module = "contract"
    _model = Contract

    class Input(ContractRenewInputType):
        pass



# Generated by Django 3.0.3 on 2020-12-28 13:10

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('policy', '0001_initial'),
        ('contract', '0006_auto_20201228_1218'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contractcontributionplandetails',
            name='policy',
            field=models.ForeignKey(db_column='PolicyID', on_delete=django.db.models.deletion.DO_NOTHING, to='policy.Policy'),
        ),
        migrations.AlterField(
            model_name='historicalcontractcontributionplandetails',
            name='policy',
            field=models.ForeignKey(blank=True, db_column='PolicyID', db_constraint=False, null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='policy.Policy'),
        ),
    ]



# Generated by Django 3.0.3 on 2021-02-08 15:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contribution', '0001_initial'),
        ('contract', '0015_auto_20210208_1506'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contractcontributionplandetails',
            name='contribution',
            field=models.ForeignKey(blank=True, db_column='ContributionId', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='contract_contribution_plan_details', to='contribution.Premium'),
        ),
    ]



# Generated by Django 3.0.3 on 2020-12-22 10:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0002_auto_20201221_1547'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contract',
            name='code',
            field=models.CharField(db_column='Code', max_length=64),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='code',
            field=models.CharField(db_column='Code', max_length=64),
        ),
    ]



import datetime

from calculation.services import (
    get_linked_class,
    get_parameters,
    get_rule_details,
    get_rule_name,
)
from contribution_plan.tests.helpers import (
    create_test_contribution_plan,
    create_test_contribution_plan_bundle,
    create_test_contribution_plan_bundle_details,
)
from core.models import User
from core.test_helpers import create_test_interactive_user
from django.test import TestCase
from policy.test_helpers import create_test_policy
from policyholder.tests.helpers import (
    create_test_policy_holder,
    create_test_policy_holder_insuree,
)

from contract.models import Contract, ContractContributionPlanDetails, ContractDetails
from contract.services import Contract as ContractService
from contract.services import (
    ContractContributionPlanDetails as ContractContributionPlanDetailsService,
)
from contract.services import ContractDetails as ContractDetailsService
from contract.services import subtract_date_ranges


class ServiceTestContract(TestCase):
    user = None

    @classmethod
    def setUpClass(cls):
        super(ServiceTestContract, cls).setUpClass()
        cls.user = User.objects.filter(username="admin").first()
        if not cls.user:
            cls.user = create_test_interactive_user(
                username="admin", password="S/pe®Pąßw0rd™"
            )
        cls.contract_service = ContractService(cls.user)
        cls.contract_details_service = ContractDetailsService(cls.user)
        cls.contract_contribution_plan_details_service = (
            ContractContributionPlanDetailsService(cls.user)
        )
        # some test data so as to created contract properly
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 5
        cls.policy_holder = create_test_policy_holder()
        cls.policy_holder2 = create_test_policy_holder()
        cls.time_stamp = datetime.datetime.now()
        cls.date_from = str((cls.time_stamp + datetime.timedelta(days=30)).date())
        cls.date_to = str((cls.time_stamp + datetime.timedelta(days=60)).date())
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()
        cls.contribution_plan = create_test_contribution_plan(
            custom_props={"json_ext": {"calculation_rule": {"rate": cls.rate}}}
        )
        cls.contribution_plan_bundle_details = (
            create_test_contribution_plan_bundle_details(
                contribution_plan=cls.contribution_plan,
                contribution_plan_bundle=cls.contribution_plan_bundle,
            )
        )
        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            ph_insuree = create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={
                    "last_policy": None,
                    "json_ext": {"calculation_rule": {"income": cls.income}},
                },
            )
            create_test_policy(
                cls.contribution_plan.benefit_plan,
                ph_insuree.insuree,
                custom_props={
                    "start_date": datetime.datetime(2016, 3, 1),
                    "expiry_date": datetime.datetime(2021, 7, 1),
                },
            )

    def test_contract_create_without_policy_holder(self):
        contract = {
            "code": "AAAAAA",
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        # tear down the test data
        # Contract.objects.filter(id=response["data"]["id"]).delete()
        self.assertEqual(
            (
                True,
                "Ok",
                "",
                "AAAAAA",
                0,
                None,
            ),
            (
                response["success"],
                response["message"],
                response["detail"],
                response["data"]["code"],
                response["data"]["amendment"],
                response["data"]["amount_notified"],
            ),
        )

    def test_contract_create_with_policy_holder(self):
        contract = {
            "code": "TESTONE",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        # tear down the test data
        # ContractDetails.objects.filter(contract_id=response["data"]["id"]).delete()
        # Contract.objects.filter(id=response["data"]["id"]).delete()
        self.assertEqual(
            (
                True,
                "Ok",
                "",
                "TESTONE",
                0,
            ),
            (
                response["success"],
                response["message"],
                response["detail"],
                response["data"]["code"],
                response["data"]["amendment"],
            ),
        )

    def test_contract_create_update_delete_with_policy_holder(self):
        contract = {
            "code": "CTSV",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
            "payment_reference": "payment_one xxxxxxxx",
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.update(contract)
        updated_payment_reference = response["data"]["payment_reference"]

        contract = {
            "id": contract_id,
        }
        response = self.contract_service.delete(contract)
        is_deleted = response["success"]

        # tear down the test data
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            ("payment_one xxxxxxxx", True), (updated_payment_reference, is_deleted)
        )

    def test_contract_create_update_failed_ph(self):
        contract = {
            "code": "CSTG",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }

        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
            "policy_holder_id": str(self.policy_holder2.id),
        }
        response = self.contract_service.update(contract)
        failed = response["detail"]

        # tear down the test data
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            "ContractUpdateError: You cannot update already set PolicyHolder in Contract!",
            failed,
        )

    def test_contract_create_submit_fail_scenarios(self):
        contract = {
            "code": "MTD",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }

        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract_created = Contract.objects.filter(id=contract_id).first()
        contract_created.state = Contract.STATE_EXECUTABLE
        contract_created.save(username="admin")

        contract = {
            "id": contract_id,
        }

        response = self.contract_service.submit(contract)
        result_message = response["detail"]
        expected_message = "ContractUpdateError: The contract cannot be submitted because of current state!"

        contract_created.state = Contract.STATE_NEGOTIABLE
        contract_created.save(username="admin")

        response = self.contract_service.submit(contract)
        result_message2 = response["detail"]
        expected_message2 = (
            "ContractUpdateError: The contract has been already submitted!"
        )

        contract_created.policy_holder = None
        contract_created.save(username="admin")

        response = self.contract_service.submit(contract)
        result_message3 = response["detail"]
        expected_message3 = (
            "ContractUpdateError: The contract does not contain PolicyHolder!"
        )

        # tear down the test data
        list_cd = list(
            ContractDetails.objects.filter(contract_id=contract_id).values(
                "id", "json_ext"
            )
        )
        self.assertNotEqual(list_cd, [])
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(
            (expected_message, expected_message2, expected_message3),
            (result_message, result_message2, result_message3),
        )

    def test_contract_create_submit_counter(self):
        contract = {
            "code": "SUR",
            "policy_holder_id": str(self.policy_holder.id),
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])

        contract = {
            "id": contract_id,
        }
        self.contract_service.submit(contract)
        response = self.contract_service.counter(contract)
        result_state = response["data"]["state"]
        expected_state = 11

        # tear down the test data
        list_cd = list(
            ContractDetails.objects.filter(contract_id=contract_id).values("id")
        )
        self.assertNotEqual(list_cd, [])
        # for cd in list_cd:
        #    ccpd = ContractContributionPlanDetails.objects.filter(contract_details__id=f"{cd['id']}").delete()
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(expected_state, result_state)

    def test_contract_create_submit(self):

        contract = {
            "code": "TESTCON",
            "policy_holder_id": str(self.policy_holder.id),
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])
        contract = {
            "id": contract_id,
        }
        response = self.contract_service.submit(contract)
        expected_state = 4
        result_state = response["data"]["state"]
        # tear down the test data
        list_cd = list(
            ContractDetails.objects.filter(contract_id=contract_id).values("id")
        )
        for cd in list_cd:
            ContractContributionPlanDetails.objects.filter(
                contract_details__id=f"{cd['id']}"
            ).delete()
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()
        self.assertEqual(expected_state, result_state)

    def test_contract_create_cd_from_phi(self):

        ph_insuree2 = create_test_policy_holder_insuree(
            policy_holder=self.policy_holder,
            contribution_plan_bundle=self.contribution_plan_bundle,
            custom_props={
                "last_policy": None,
                "json_ext": {"calculation_rule": {"income": 400}},
            },
        )
        contract = {
            "code": "MTEST-1",
            "policy_holder_id": str(self.policy_holder.id),
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }
        response = self.contract_service.create(contract)
        contract_id = str(response["data"]["id"])
        contract = Contract.objects.get(id=contract_id)

        response = self.contract_details_service.get_details_from_ph_insuree(
            contract=contract, ph_insuree=ph_insuree2.id
        )
        # tear down the test data
        # ContractDetails.objects.filter(contract_id=contract_id).delete()
        # Contract.objects.filter(id=contract_id).delete()

        self.assertEqual(True, response["success"])

    def test_date_subtract(self):
        date_range = (
            datetime.date(2024, 1, 1),
            datetime.date(2024, 12, 31),
        )
        date_ranges = [
            (datetime.date(2024, 2, 1), datetime.date(2024, 3, 15)),
            (datetime.date(2024, 5, 1), datetime.date(2024, 6, 30)),
            (datetime.date(2024, 8, 1), datetime.date(2024, 9, 30)),
        ]

        result = subtract_date_ranges(date_range, date_ranges)

        per_1 = (datetime.date(2024, 1, 1), datetime.date(2024, 2, 1))
        per_2 = (datetime.date(2024, 3, 15), datetime.date(2024, 5, 1))
        per_3 = (datetime.date(2024, 6, 30), datetime.date(2024, 8, 1))
        per_4 = (datetime.date(2024, 9, 30), datetime.date(2024, 12, 31))

        self.assertTrue(per_1 in result)
        self.assertTrue(per_2 in result)
        self.assertTrue(per_3 in result)
        self.assertTrue(per_4 in result)


class CalculationContractTest(TestCase):
    user = None

    @classmethod
    def setUpClass(cls):
        super(CalculationContractTest, cls).setUpClass()
        cls.user = User.objects.filter(username="admin").first()
        if not cls.user:
            cls.user = create_test_interactive_user(
                username="admin", password="S/pe®Pąßw0rd™"
            )
        cls.contract_service = ContractService(cls.user)
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 5
        cls.policy_holder = create_test_policy_holder()
        cls.time_stamp = datetime.datetime.now()
        cls.date_from = str((cls.time_stamp + datetime.timedelta(days=30)).date())
        cls.date_to = str((cls.time_stamp + datetime.timedelta(days=60)).date())
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()
        cls.contribution_plan = create_test_contribution_plan(
            custom_props={"json_ext": {"calculation_rule": {"rate": cls.rate}}}
        )
        cls.contribution_plan_bundle_details = (
            create_test_contribution_plan_bundle_details(
                contribution_plan=cls.contribution_plan,
                contribution_plan_bundle=cls.contribution_plan_bundle,
            )
        )

        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={
                    "last_policy": None,
                    "json_ext": {"calculation_rule": {"income": cls.income}},
                },
            )

    def test_get_rule_name(self):
        class_name = "ContractDetails"
        result = get_rule_name(class_name=class_name)
        self.assertEqual("ContributionValuationRule", result[0].__name__)

    def test_get_rule_not_existing(self):
        class_name = "xxxxxxxxxxxxxxxxxxx"
        result = get_rule_name(class_name=class_name)
        self.assertEqual([], result)

    def test_get_rule_details(self):
        class_name = "PolicyHolderInsuree"
        class_name2 = "ContributionPlan"
        result = get_rule_details(class_name=class_name)
        result2 = get_rule_details(class_name=class_name2)
        result_param = [param["name"] for param in result[class_name]]
        result2_param = [param["name"] for param in result2[class_name2]]
        self.assertEqual(
            (class_name, class_name2, ["income"], ["rate", "includeFamily"]),
            (
                list(result.keys())[0],
                list(result2.keys())[0],
                result_param,
                result2_param,
            ),
        )

    def test_get_rule_details_not_existing(self):
        class_name = "xxxxxxxxxxxxxxx"
        result = get_rule_details(class_name=class_name)
        self.assertEqual({}, result)

    def test_get_linked_class_empty(self):
        result = get_linked_class()
        self.assertEqual(["Calculation"], result)

    def test_get_linked_class(self):
        result = get_linked_class(["PolicyHolderInsuree"])
        self.assertEqual(
            sorted(
                [
                    "PolicyHolder",
                    "Insuree",
                    "ContributionPlanBundle",
                    "Policy",
                    "Calculation",
                ]
            ),
            sorted(result),
        )

    def test_get_param_and_amount_calculation(self):
        # create contract to test 1st calculation rule contribution valuation
        # test case - create contract and check if amount is calcutated properly
        # and test if on instance of contract details the proper params is showed
        # by getting param name
        contract = {
            "code": "CALTEST",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": self.date_from,
            "date_valid_to": self.date_to,
        }

        response = self.contract_service.create(contract)
        self.assertTrue(response["success"])
        # after creating contract - we can get contract details so as to get params
        # run calculation rules etc
        cd = ContractDetails.objects.filter(contract_id=response["data"]["id"]).first()
        # c = Contract.objects.filter(id=response["data"]["id"]).first()
        result_params = get_parameters("PolicyHolderInsuree", cd)

        # tear down the contract test data
        # cd.delete()
        # c.delete()

        # we want to assert name of param related to the contract details (should be 'income')
        # and also the value of contract (all contributions)
        # for that test should be income=500, 5 insurees, default rate=5 %
        # income*rate*number of contributions = according to Contribution Valuation Rule
        self.assertEqual(
            ("income", self.income * (float(self.rate / 100)) * self.number_of_insuree),
            (result_params[0]["name"], response["data"]["amount_notified"]),
        )

    def test_no_deatils(self):
        # create contract to test 1st calculation rule contribution valuation
        # test case - create contract and check if amount is calcutated properly
        # and test if on instance of contract details the proper params is showed
        # by getting param name
        contract = {
            "code": "CALTEST",
            "policy_holder_id": self.policy_holder.id,
            "date_valid_from": datetime.date(2017, 1, 1),
            "date_valid_to": datetime.date(2017, 1, 31),
        }

        response = self.contract_service.create(contract)
        self.assertTrue(response["success"])
        cd = ContractDetails.objects.filter(contract_id=response["data"]["id"]).first()
        self.assertIsNone(
            cd,
            "the contract should not have any detail as there is no insuree valid at this time",
        )



# Generated by Django 4.2.16 on 2024-10-30 08:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("policy", "0010_merge_20240912_1248"),
        ("contract", "0022_add_admin_rights"),
    ]

    operations = [
        migrations.AddField(
            model_name="contractcontributionplandetails",
            name="amount",
            field=models.DecimalField(
                blank=True,
                db_column="Amount",
                decimal_places=2,
                max_digits=18,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalcontractcontributionplandetails",
            name="amount",
            field=models.DecimalField(
                blank=True,
                db_column="Amount",
                decimal_places=2,
                max_digits=18,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="contractcontributionplandetails",
            name="policy",
            field=models.ForeignKey(
                blank=True,
                db_column="PolicyID",
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                to="policy.policy",
            ),
        ),
    ]



import base64
import datetime
from unittest import mock
from uuid import UUID

import graphene
from django.test import TestCase
from graphene import Schema
from graphene.test import Client

from contract import schema as contract_schema
from contract.tests.helpers import (
    create_test_contract,
    create_test_contract_contribution_plan_details,
    create_test_contract_details,
)


class ContractQueryTest(TestCase):
    class BaseTestContext:
        user = mock.Mock(is_anonymous=False)
        user.has_perm = mock.MagicMock(return_value=False)

    class AnonymousUserContext:
        user = mock.Mock(is_anonymous=True)

    @classmethod
    def setUpClass(cls):
        super(ContractQueryTest, cls).setUpClass()
        cls.date_created = datetime.datetime.now()
        cls.test_contract = create_test_contract(
            custom_props={
                "code": "testContract-" + str(cls.date_created),
                "payment_reference": "payment reference" + str(cls.date_created),
                "amount_due": 450.99,
                "date_valid_from": datetime.date(2020, 1, 1),
                "amendment": 1,
            }
        )
        cls.test_contract_details = create_test_contract_details(
            contract=cls.test_contract
        )
        cls.test_contract_contribution_plan_details = (
            create_test_contract_contribution_plan_details(
                contract_details=cls.test_contract_details
            )
        )

        cls.schema = Schema(
            query=contract_schema.Query,
        )

        cls.graph_client = Client(cls.schema)

    def test_find_contract_existing(self):
        id = self.test_contract.id
        result = self.find_by_id_query("contract", id)
        converted_id = (
            base64.b64decode(result[0]["node"]["id"]).decode("utf-8").split(":")[1]
        )
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_details_existing(self):
        id = self.test_contract_details.id
        result = self.find_by_id_query("contractDetails", id)
        converted_id = (
            base64.b64decode(result[0]["node"]["id"]).decode("utf-8").split(":")[1]
        )
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_existing(self):
        id = self.test_contract_contribution_plan_details.id
        result = self.find_by_id_query("contractContributionPlanDetails", id)
        converted_id = (
            base64.b64decode(result[0]["node"]["id"]).decode("utf-8").split(":")[1]
        )
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_params(self):
        expected = self.test_contract
        params = {
            "version": expected.version,
            "isDeleted": True if expected.is_deleted else False,
            "code": expected.code,
        }
        result = self.find_by_exact_attributes_query("contract", params)
        self.assertDictEqual(result[0]["node"], params)

    def test_find_contract_details_by_contract(self):
        details_contract_id = self.test_contract_details.contract.id
        id = self.test_contract_details.id
        query = f"""
        {{
            contractDetails(
                contract_Id: "{details_contract_id}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_details_by_id_and_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract_details.id
        query = f"""
        {{
            contractDetails(
                id: "{id}"
                insuree_Uuid: "{details_insuree_uuid}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code(self):
        code = self.test_contract.code
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code: "{code}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_payment_reference(self):
        payment_reference = self.test_contract.payment_reference
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                paymentReference: "{payment_reference}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code:" {code}", amountDue: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due_greater(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due - 100
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code:" {code}", amountDue_Gte: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amount_due_greater_equal(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code:" {code}", amountDue_Gte: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_amound_due_lt(self):
        code = self.test_contract.code
        amount_due = self.test_contract.amount_due
        # id = self.test_contract.id
        query = f"""
        {{
            contract(
                code:" {code}", amountDue_Lt: {amount_due}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"]
        self.assertEqual(len(result), 0)

    def test_find_contract_by_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                insuree: "{details_insuree_uuid}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_id_and_state(self):
        details_contract_state = self.test_contract.state
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                id: "{id}"
                state: {details_contract_state}) {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_contract_details_and_contribution_plan(
        self,
    ):
        details_contract_details_id = (
            self.test_contract_contribution_plan_details.contract_details.id
        )
        details_contribution_plan_id = (
            self.test_contract_contribution_plan_details.contribution_plan.id
        )
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                contractDetails_Id: "{details_contract_details_id}",
                contributionPlan_Id: "{details_contribution_plan_id}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_date_valid_from_gte(self):
        code = self.test_contract.code
        date_valid_from = str(self.test_contract.date_valid_from.date()) + "T00:00:00"
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code: "{code}", dateValidFrom_Gte: "{date_valid_from}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_contains(self):
        date = self.date_created
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code_Icontains: "{date}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_code_amendment(self):
        code = self.test_contract.code
        amendment = self.test_contract.amendment
        id = self.test_contract.id
        query = f"""
        {{
            contract(
                code: "{code}", amendment: {amendment}){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_by_contract_code_start_with(self):
        query = f"""
        {{
            contract(
                code_Istartswith: "{'testContract-'}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contract"]["edges"]
        self.assertGreater(len(result), 0)

    def test_find_contracts_details_by_date_created(self):
        date_created = str(self.test_contract_details.date_created).replace(" ", "T")
        id = self.test_contract_details.id
        query = f"""
        {{
            contractDetails(
                dateCreated: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contracts_details_by_date_updated(self):
        date_updated = str(self.test_contract_details.date_updated).replace(" ", "T")
        id = self.test_contract_details.id
        query = f"""
        {{
            contractDetails(
                dateUpdated: "{date_updated}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created(self):
        date_created = str(
            self.test_contract_contribution_plan_details.date_created
        ).replace(" ", "T")
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                dateCreated: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created_gte(self):
        date_created = str(
            self.test_contract_contribution_plan_details.date_created
        ).replace(" ", "T")
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                dateCreated_Gte: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_date_created_gt(self):
        date_created = str(
            self.test_contract_contribution_plan_details.date_created
        ).replace(" ", "T")
        # id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                dateCreated_Gt: "{date_created}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"]
        self.assertEqual(len(result), 0)

    def test_find_contract_contribution_plan_details_by_date_updated(self):
        date_updated = str(
            self.test_contract_contribution_plan_details.date_updated
        ).replace(" ", "T")
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                dateUpdated: "{date_updated}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_contribution(self):
        details_contribution_uuid = (
            self.test_contract_contribution_plan_details.contribution.uuid
        )
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Uuid: "{details_contribution_uuid}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_amount_in_contribution_gte(
        self,
    ):
        details_contribution_amount = (
            self.test_contract_contribution_plan_details.contribution.amount
        )
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Amount_Gte: "{details_contribution_amount}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def test_find_contract_contribution_plan_details_by_id_and_amount_in_contribution_gt(
        self,
    ):
        details_contribution_amount = (
            self.test_contract_contribution_plan_details.contribution.amount
        )
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                id: "{id}"
                contribution_Amount_Gt: "{details_contribution_amount}"){{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"]
        self.assertEqual(len(result), 0)

    def test_find_contract_contribution_plan_details_by_insuree(self):
        details_insuree_uuid = self.test_contract_details.insuree.uuid
        id = self.test_contract_contribution_plan_details.id
        query = f"""
        {{
            contractContributionPlanDetails(
                insuree: "{details_insuree_uuid}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query)
        result = query_result["contractContributionPlanDetails"]["edges"][0]["node"]
        converted_id = base64.b64decode(result["id"]).decode("utf-8").split(":")[1]
        self.assertEqual(UUID(converted_id), id)

    def find_by_id_query(self, query_type, id, context=None):
        query = f"""
        {{
            {query_type}(id:"{id}") {{
                totalCount
                edges {{
                  node {{
                    id
                  }}
                  cursor
                }}
          }}
        }}
        """

        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]["edges"]

        if len(records) > 1:
            raise ValueError(f"Ambiguous id {id} for query {query_type}")

        return records

    def find_by_exact_attributes_query(self, query_type, params, context=None):
        node_content_str = "\n".join(params.keys())
        query = f"""
        {{
            {query_type}({self.build_params(params)}) {{
                totalCount
                edges {{
                  node {{
                    {node_content_str}
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]["edges"]
        return records

    def execute_query(self, query, context=None):
        if context is None:
            context = self.BaseTestContext()

        query_result = self.graph_client.execute(query, context=context)
        query_data = query_result["data"]
        return query_data

    def build_params(self, params):
        def wrap_arg(v):
            if isinstance(v, str):
                return f'"{v}"'
            if isinstance(v, bool):
                return str(v).lower()
            if isinstance(v, datetime.date):
                return graphene.DateTime.serialize(
                    datetime.datetime.fromordinal(v.toordinal())
                )
            return v

        params_as_args = [f"{k}:{wrap_arg(v)}" for k, v in params.items()]
        return ", ".join(params_as_args)



import logging

from core.models import User

from contract.models import Contract, ContractContributionPlanDetails
from contract.services import Contract as ContractService
from contract.services import ContractToInvoiceService

logger = logging.getLogger(__name__)


def approve_contracts(user_id, contracts):
    output = []
    user = User.objects.get(id=user_id)
    contract_service = ContractService(user=user)
    for contract in contracts:
        output.append(contract_service.approve(contract={"id": contract}))
    return output


def counter_contracts(user_id, contracts):
    output = []
    user = User.objects.get(id=user_id)
    contract_service = ContractService(user=user)
    for contract in contracts:
        output.append(contract_service.counter(contract={"id": contract}))
    return output


def create_invoice_from_contracts(user_id, contracts):
    output = []
    user = User.objects.get(id=user_id)
    contract_service = ContractToInvoiceService(user=user)
    for contract in contracts:
        contract_instance = Contract.objects.filter(id=contract)
        if contract_instance:
            contract_instance = contract_instance.first()
            ccpd_list = ContractContributionPlanDetails.objects.filter(
                contract_details__contract=contract_instance
            )
            output.append(
                contract_service.create_invoice(
                    instance=contract_instance,
                    convert_to="InvoiceLine",
                    user=user,
                    ccpd_list=ccpd_list,
                )
            )
    return output



# Generated by Django 3.0.3 on 2021-01-19 12:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0009_auto_20210118_1427'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contractcontributionplandetails',
            name='contract_details',
            field=models.ForeignKey(db_column='ContractDetailsUUID', on_delete=django.db.models.deletion.CASCADE, to='contract.ContractDetails'),
        ),
        migrations.AlterField(
            model_name='contractdetails',
            name='contract',
            field=models.ForeignKey(db_column='ContractUUID', on_delete=django.db.models.deletion.CASCADE, to='contract.Contract'),
        ),
    ]



from core.gql.gql_mutations import DeleteInputType
from core.gql.gql_mutations.base_mutation import (
    BaseDeleteMutation,
    BaseHistoryModelCreateMutationMixin,
    BaseHistoryModelDeleteMutationMixin,
    BaseHistoryModelUpdateMutationMixin,
    BaseMutation,
)

from contract.gql.gql_mutations.input_types import (
    ContractDetailsCreateFromInsureeInputType,
    ContractDetailsCreateInputType,
    ContractDetailsUpdateInputType,
)
from contract.models import ContractDetails, ContractDetailsMutation

from .mutations import ContractDetailsFromPHInsureeMutationMixin


class CreateContractDetailsMutation(BaseHistoryModelCreateMutationMixin, BaseMutation):
    _mutation_class = "ContractDetailsMutation"
    _mutation_module = "contract"
    _model = ContractDetails

    @classmethod
    def _mutate(cls, user, **data):
        client_mutation_id = data.get("client_mutation_id")
        if "client_mutation_id" in data:
            data.pop("client_mutation_id")
        if "client_mutation_label" in data:
            data.pop("client_mutation_label")
        contract_detail = cls.create_object(user=user, object_data=data)
        ContractDetailsMutation.object_mutated(
            user, client_mutation_id=client_mutation_id, contract_detail=contract_detail
        )
        return None

    class Input(ContractDetailsCreateInputType):
        pass


class UpdateContractDetailsMutation(BaseHistoryModelUpdateMutationMixin, BaseMutation):
    _mutation_class = "ContractDetailsMutation"
    _mutation_module = "contract"
    _model = ContractDetails

    class Input(ContractDetailsUpdateInputType):
        pass


class DeleteContractDetailsMutation(
    BaseHistoryModelDeleteMutationMixin, BaseDeleteMutation
):
    _mutation_class = "ContractDetailsMutation"
    _mutation_module = "contract"
    _model = ContractDetails

    class Input(DeleteInputType):
        pass


class CreateContractDetailByPolicyHolderInsureeMutation(
    ContractDetailsFromPHInsureeMutationMixin, BaseMutation
):
    _mutation_class = "CreateContractDetailByPolicyHolderInsureetMutation"
    _mutation_module = "contract"
    _model = ContractDetails

    class Input(ContractDetailsCreateFromInsureeInputType):
        pass



import logging

from django.db import migrations

from core.utils import insert_role_right_for_system

logger = logging.getLogger(__name__)


def add_rights(apps, schema_editor):
    insert_role_right_for_system(256, 152109, apps)  # contract approve


class Migration(migrations.Migration):
    dependencies = [
        ('contract', '0018_approve_ask_for_change_perms')
    ]

    operations = [
        migrations.RunPython(add_rights),
    ]



# Generated by Django 3.2.16 on 2023-01-26 09:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0019_amend_perms'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='historicalcontract',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical contract', 'verbose_name_plural': 'historical contracts'},
        ),
        migrations.AlterModelOptions(
            name='historicalcontractcontributionplandetails',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical contract contribution plan details', 'verbose_name_plural': 'historical contract contribution plan detailss'},
        ),
        migrations.AlterModelOptions(
            name='historicalcontractdetails',
            options={'get_latest_by': ('history_date', 'history_id'), 'ordering': ('-history_date', '-history_id'), 'verbose_name': 'historical contract details', 'verbose_name_plural': 'historical contract detailss'},
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='history_date',
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractcontributionplandetails',
            name='history_date',
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractdetails',
            name='history_date',
            field=models.DateTimeField(db_index=True),
        ),
    ]



from functools import lru_cache

from django.test import TestCase

from ..models import (
    ContractContributionPlanDetails,
    Contract,
    ContractDetails,
)
from .helpers import (
    create_test_contract_contribution_plan_details,
    create_test_contract_details,
    create_test_contract,
)
from insuree.test_helpers import (
    create_test_insuree,
)
from policyholder.tests.helpers import (
    create_test_policy_holder
)
from contribution_plan.tests.helpers import (
    create_test_contribution_plan_bundle
)
from contribution_plan.tests.helpers import create_test_contribution_plan


class HelpersTest(TestCase):
    """
    Class to check whether the helper methods responsible for creating test data work correctly.
    """

    def test_create_test_contract(self):
        contract = self.__create_test_contract()
        db_contract = Contract.objects.filter(id=contract.uuid).first()
        self.assertEqual(db_contract, contract, "Failed to create contract in helper")

    def test_create_test_contract_custom(self):
        contract = self.__create_test_contract(custom=True)
        db_contract = Contract.objects.filter(id=contract.uuid).first()
        params = self.__custom_contract_params
        self.assertEqual(db_contract.version, params["version"])
        self.assertEqual(db_contract.policy_holder, params["policy_holder"])
        self.assertEqual(db_contract.state, params["state"])

    def test_create_test_contract_details(self):
        contract_details = self.__create_test_contract_details()
        db_contract_details = ContractDetails.objects.filter(
            id=contract_details.uuid
        ).first()
        self.assertEqual(
            db_contract_details,
            contract_details,
            "Failed to create contract details in helper",
        )

    def test_create_test_contract_details_custom(self):
        contract_details = self.__create_test_contract_details(custom=True)
        db_contract_details = ContractDetails.objects.filter(
            id=contract_details.uuid
        ).first()
        params = self.__custom_contract_details_params

        self.assertEqual(db_contract_details.contract, params["contract"])
        self.assertEqual(db_contract_details.insuree, params["insuree"])
        self.assertEqual(
            db_contract_details.contribution_plan_bundle,
            params["contribution_plan_bundle"],
        )

    def test_create_test_contract_contribution_plan_details(self):
        contract_contribution_plan_details = (
            self.__create_test_contract_contribution_plan_details()
        )
        db_contract_contribution_plan_details = (
            ContractContributionPlanDetails.objects.filter(
                id=contract_contribution_plan_details.uuid
            ).first()
        )
        self.assertEqual(
            db_contract_contribution_plan_details,
            contract_contribution_plan_details,
            "Failed to create contract contribution plan details in helper",
        )

    def test_create_test_contract_contribution_plan_details_custom(self):
        contract_contribution_plan_details = (
            self.__create_test_contract_contribution_plan_details(custom=True)
        )
        db_contract_contribution_plan_details = (
            ContractContributionPlanDetails.objects.filter(
                id=contract_contribution_plan_details.uuid
            ).first()
        )
        params = self.__custom_contract_contribution_plan_details_params
        self.assertEqual(
            db_contract_contribution_plan_details.contribution_plan,
            params["contribution_plan"],
        )
        self.assertEqual(
            db_contract_contribution_plan_details.contract_details,
            params["contract_details"],
        )

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_params(self):
        return {
            "version": 2,
            "policy_holder": create_test_policy_holder(custom_props={"version": 2}),
            "state": 1,
        }

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_details_params(self):
        return {
            "contract": self.__create_test_contract(True),
            "contribution_plan_bundle": create_test_contribution_plan_bundle(
                custom_props={"version": 3}
            ),
            "insuree": create_test_insuree(),
        }

    @property
    @lru_cache(maxsize=2)
    def __custom_contract_contribution_plan_details_params(self):
        return {
            "contribution_plan": create_test_contribution_plan(
                custom_props={"version": 2}
            ),
            "contract_details": self.__create_test_contract_details(True),
        }

    def __create_test_instance(self, function, **kwargs):
        if kwargs:
            return function(**kwargs)
        else:
            return function()

    def __create_test_contract(self, custom=False):
        custom_params = self.__custom_contract_params if custom else {}
        return self.__create_test_instance(
            create_test_contract, custom_props=custom_params
        )

    def __create_test_contract_details(self, custom=False):
        custom_params = self.__custom_contract_details_params if custom else {}
        return self.__create_test_instance(
            create_test_contract_details, custom_props=custom_params
        )

    def __create_test_contract_contribution_plan_details(self, custom=False):
        custom_params = (
            self.__custom_contract_contribution_plan_details_params if custom else {}
        )
        return self.__create_test_instance(
            create_test_contract_contribution_plan_details, custom_props=custom_params
        )



import base64
import datetime
import json
import uuid
from unittest import mock

import graphene
from contribution_plan.tests.helpers import (
    create_test_contribution_plan,
    create_test_contribution_plan_bundle,
    create_test_contribution_plan_bundle_details,
)
from core.models import User
from core.models.openimis_graphql_test_case import openIMISGraphQLTestCase
from core.test_helpers import create_test_interactive_user
from django.conf import settings
from graphene import Schema
from graphene.test import Client
from graphql_jwt.shortcuts import get_token
from policy.test_helpers import create_test_policy
from policyholder.tests.helpers import (
    create_test_policy_holder,
    create_test_policy_holder_insuree,
)
from contract import schema as contract_schema
from contract.models import Contract, ContractContributionPlanDetails
from contract.services import subtract_date_ranges
from contract.signals import append_contract_filter
from payment.models import Payment
from insuree.models import InsureePolicy


class MutationTestContract(openIMISGraphQLTestCase):
    GRAPHQL_URL = f"/{settings.SITE_ROOT()}graphql"
    # This is required by some version of graphene but is never used. It should be set to the schema but the import
    # is shown as an error in the IDE, so leaving it as True.
    GRAPHQL_SCHEMA = True
    admin_user = None
    schema = Schema(query=contract_schema.Query)

    class BaseTestContext:
        def __init__(self, user):
            self.user = user

    class AnonymousUserContext:
        user = mock.Mock(is_anonymous=True)

    @classmethod
    def setUpClass(cls):
        super(MutationTestContract, cls).setUpClass()
        cls.user = User.objects.filter(username="admin", i_user__isnull=False).first()
        if not cls.user:
            cls.user = create_test_interactive_user(username="admin")
        # some test data so as to created contract properly
        cls.user_token = get_token(cls.user, cls.BaseTestContext(user=cls.user))
        cls.income = 500
        cls.rate = 5
        cls.number_of_insuree = 2
        cls.policy_holder = create_test_policy_holder()
        cls.policy_holder2 = create_test_policy_holder()
        cls.time_stamp = datetime.datetime.now()
        cls.date_from = str((cls.time_stamp + datetime.timedelta(days=30)).date())
        cls.date_to = str((cls.time_stamp + datetime.timedelta(days=60)).date())
        # create contribution plans etc
        cls.contribution_plan_bundle = create_test_contribution_plan_bundle()

        cls.contribution_plan = create_test_contribution_plan(
            custom_props={
                "json_ext": {"calculation_rule": {"rate": cls.rate}},
                "date_valid_from": "2010-01-01",
                "date_valid_to": "2020-01-01",
            }
        )
        cls.contribution_plan_old = create_test_contribution_plan(
            custom_props={
                "json_ext": {"calculation_rule": {"rate": cls.rate}},
                "date_valid_from": "2020-01-01",
                "replacement_uuid": cls.contribution_plan.id
            }
        )

        cls.contribution_plan_bundle_details = create_test_contribution_plan_bundle_details(
            contribution_plan=cls.contribution_plan,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )
        cls.contribution_plan_bundle_details = create_test_contribution_plan_bundle_details(
            contribution_plan=cls.contribution_plan_old,
            contribution_plan_bundle=cls.contribution_plan_bundle
        )
        # create policy holder insuree for that test policy holder
        for i in range(0, cls.number_of_insuree):
            ph_insuree = create_test_policy_holder_insuree(
                policy_holder=cls.policy_holder,
                contribution_plan_bundle=cls.contribution_plan_bundle,
                custom_props={
                    "last_policy": None,
                    "json_ext": {"calculation_rule": {"income": cls.income}},
                },
            )
            create_test_policy(
                cls.contribution_plan.benefit_plan,
                ph_insuree.insuree,
                custom_props={
                    "start_date": datetime.datetime(2016, 3, 1),
                    "expiry_date": datetime.datetime(2021, 7, 1),
                },
            )

        cls.policy_holder_insuree_old = create_test_policy_holder_insuree(
            policy_holder=cls.policy_holder,
            contribution_plan_bundle=cls.contribution_plan_bundle,
            custom_props={
                "json_ext": {"calculation_rule": {"income": cls.income}},
                "date_valid_from": "2010-01-01",
                "date_valid_to": "2020-01-01",
            }
        )

        cls.policy_holder_insuree = create_test_policy_holder_insuree(
            policy_holder=cls.policy_holder,
            contribution_plan_bundle=cls.contribution_plan_bundle,
            custom_props={
                "json_ext": {"calculation_rule": {"income": cls.income}},
                "date_valid_from": "2020-01-01",
                "replacement_uuid": cls.contribution_plan.id
            }
        )

        cls.policy_holder_insuree2 = create_test_policy_holder_insuree(
            policy_holder=cls.policy_holder,
            contribution_plan_bundle=cls.contribution_plan_bundle,
            custom_props={
                "json_ext": {"calculation_rule": {"income": cls.income}},
            }
        )

        cls.schema = Schema(
            query=contract_schema.Query, mutation=contract_schema.Mutation
        )
        cls.graph_client = Client(cls.schema)

    def test_mutation_contract_create_without_policy_holder(self):

        input_param = {
            "code": "XYZ:" + str(self.time_stamp),
            "dateValidFrom": self.date_from,
            "dateValidTo": self.date_to,
        }
        self.add_mutation("createContract", input_param, self.user_token)
        result = self.find_by_exact_attributes_query(
            "contract",
            params=input_param,
        )["edges"]
        # converted_id = base64.b64decode(result[0]['node']['id']).decode('utf-8').split(':')[1]
        # tear down the test data
        # Contract.objects.filter(id=str(converted_id)).delete()
        self.assertEqual(("XYZ:" + str(self.time_stamp),), (result[0]["node"]["code"],))

    def test_mutation_contract_create_with_policy_holder(self):
        input_param = {
            "code": "XYZ:" + str(self.time_stamp),
            "policyHolderId": str(self.policy_holder.id),
            "clientMutationId": str(uuid.uuid4()),
            "dateValidFrom": self.date_from,
            "dateValidTo": self.date_to,
        }
        content = self.send_mutation("createContract", input_param, self.user_token)
        self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )
        del input_param["clientMutationId"]
        result = self.find_by_exact_attributes_query(
            "contract",
            params=input_param,
        )["edges"]
        converted_id = (
            base64.b64decode(result[0]["node"]["id"]).decode("utf-8").split(":")[1]
        )
        # SUBMIT
        input_param = {"id": converted_id, "clientMutationId": str(uuid.uuid4())}
        content = self.send_mutation("submitContract", input_param, self.user_token)
        content = self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )

        # COUNTER
        input_param = {"id": converted_id, "clientMutationId": str(uuid.uuid4())}
        content = self.send_mutation("counterContract", input_param, self.user_token)

        content = self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )

        # reSUBMIT
        input_param = {"id": converted_id, "clientMutationId": str(uuid.uuid4())}
        content = self.send_mutation("submitContract", input_param, self.user_token)
        self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )
        # Approve
        input_param = {"id": converted_id, "clientMutationId": str(uuid.uuid4())}
        content = self.send_mutation("approveContract", input_param, self.user_token)

        self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )
        # Pay
        contract = Contract.objects.get(id=converted_id)
        payment = Payment.objects.filter(
            append_contract_filter(
                None,
                user=self.user,
                additional_filter={
                    'contract': contract.id
                }
            )
        ).first()
        input_param = {
            "uuid": str(payment.uuid),
            "clientMutationId": str(uuid.uuid4()),
            "receivedAmount": str(payment.expected_amount),
            "expectedAmount": str(payment.expected_amount),
            "receiptNo": "tests",
            "typeOfPayment": "C",
        }
        content = self.send_mutation("updatePayment", input_param, self.user_token)
        self.assertEqual(
            content["data"]["mutationLogs"]["edges"][0]["node"]["status"], 2
        )
        contract.refresh_from_db()
        self.assertEqual(contract.state, Contract.STATE_EFFECTIVE, 'contract not effective')
        insuree_policies = list(InsureePolicy.objects.filter(policy__in=ContractContributionPlanDetails.objects.filter(
            contract_details__contract=contract
        ).values_list('policy_id', flat=True)))

        for d in list(contract.contractdetails_set.all()):
            ips = [ip for ip in insuree_policies if ip.insuree_id == d.insuree_id]
            self.assertTrue(len(ips) > 0)
            not_covered = subtract_date_ranges(
                (contract.date_valid_from, contract.date_valid_to,),
                [(ip.effective_date, ip.expiry_date,) for ip in ips]
            )
            self.assertTrue(not_covered == [])

        # check the contract details

        query = f"""
    {{
      contractContributionPlanDetails(contractDetails_Contract_Id: "{
          str(contract.id)}",isDeleted: false,first: 10,orderBy: ["contractDetails_Insuree_Uuid"])
      {{
        totalCount

    pageInfo {{ hasNextPage, hasPreviousPage, startCursor, endCursor}}
    edges
    {{
      node
      {{
        jsonExt,contractDetails{{
            id,
            jsonExt,
            contract{{id}},
            insuree{{id, uuid, chfId, lastName, otherNames, dob}},
            contributionPlanBundle{{
                id, code, name, periodicity,
                dateValidFrom, dateValidTo,
                isDeleted, replacementUuid
            }}
        }},
        contributionPlan{{id, code, name}}
      }}
    }}
      }}
    }}
        """

        response = self.query(
            query,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.user_token}"},
        )
        content = json.loads(response.content)['data']
        self.assertEqual(len(content["contractContributionPlanDetails"]["edges"]),
                         4, "number of detail is not as expected")

        # check covered persons query

        query = f"""
    {{
      insureePolicy(additionalFilter: "{{\\"contract\\":\\"{str(contract.id)}\\"}}",first: 10,orderBy: ["insuree"])
      {{
        totalCount
        pageInfo {{ hasNextPage, hasPreviousPage, startCursor, endCursor}}
        edges
        {{
        node
        {{
            insuree{{id, uuid, chfId, lastName, otherNames, dob}}
        }}
        }}
      }}
    }}
        """
        response = self.query(
            query,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.user_token}"},
        )
        content = json.loads(response.content)['data']
        self.assertEqual(len(content["insureePolicy"]["edges"]), 4, "number of insuree Policy is not as expected")

    def find_by_id_query(self, query_type, id, context=None):
        query = f"""
        {{
            {query_type}(id:"{id}") {{
                totalCount
                edges {{
                  node {{
                    id
                    version
                  }}
                  cursor
                }}
          }}
        }}
        """

        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]["edges"]

        if len(records) > 1:
            raise ValueError(f"Ambiguous id {id} for query {query_type}")

        return records

    def find_by_exact_attributes_query(self, query_type, params, context=None):
        if "dateValidFrom" in params:
            params.pop("dateValidFrom")
        if "clientMutationId" in params:
            params.pop("clientMutationId")
        if "dateValidTo" in params:
            params.pop("dateValidTo")
        if "policyHolderId" in params:
            params.pop("policyHolderId")
        if "contributionPlanBundleId" in params:
            params.pop("contributionPlanBundleId")
        if "insureeId" in params:
            params.pop("insureeId")
        if "uuids" in params:
            params["id"] = params["uuids"][0]
            params.pop("uuids")
        node_content_str = "\n".join(params.keys()) if query_type == "contract" else ""
        query = f"""
        {{
            {query_type}({'contract_Id: "' + str(params["contractId"])
                + '", orderBy: ["-dateCreated"]' if "contractId" in params else self.build_params(params)}) {{
                totalCount
                edges {{
                  node {{
                    id
                    {node_content_str}
                    version
                    {'amountDue' if query_type == 'contract' else ''}
                    {'amountNotified' if query_type == 'contract' else ''}
                    {'amountRectified' if query_type == 'contract' else ''}
                    {'state' if query_type == 'contract' else ''}
                    {'paymentReference' if query_type == 'contract' else ''}
                  }}
                  cursor
                }}
          }}
        }}
        """
        query_result = self.execute_query(query, context=context)
        records = query_result[query_type]
        return records

    def execute_query(self, query, context=None):
        if context is None:
            context = self.BaseTestContext(self.user)

        query_result = self.graph_client.execute(query, context=context)
        query_data = query_result["data"]
        return query_data

    def add_mutation(self, mutation_type, input_params, context=None):

        if "clientMutationId" not in input_params:
            input_params["clientMutationId"] = str(uuid.uuid4())
        mutation = f"""
        mutation
        {{
            {mutation_type}(input: {{
               {self.build_params(input_params)}
            }})

          {{
            internalId
            clientMutationId
          }}
        }}
        """
        mutation_result = self.query(
            mutation,
            headers={"HTTP_AUTHORIZATION": f"Bearer {self.user_token}"},
        )
        self.assertResponseNoErrors(mutation_result)
        content = json.loads(mutation_result.content)
        return content

    def execute_mutation(self, mutation, context=None):
        if context is None:
            context = self.BaseTestContext(self.user)

        mutation_result = self.graph_client.execute(mutation, context=context)
        return mutation_result

    def build_params(self, params):
        def wrap_arg(v):
            if isinstance(v, str):
                return f'"{v}"'
            if isinstance(v, list):
                return json.dumps(v)
            if isinstance(v, bool):
                return str(v).lower()
            if isinstance(v, datetime.date):
                return graphene.DateTime.serialize(
                    datetime.datetime.fromordinal(v.toordinal())
                )
            return v

        params_as_args = [f"{k}:{wrap_arg(v)}" for k, v in params.items()]
        return ", ".join(params_as_args)



import graphene
from contribution.gql_queries import PremiumGQLType
from contribution_plan.gql.gql_types import (
    ContributionPlanBundleGQLType,
    ContributionPlanGQLType,
)
from core import ExtendedConnection, prefix_filterset
from graphene_django import DjangoObjectType
from insuree.schema import InsureeGQLType
from policyholder.gql.gql_types import PolicyHolderGQLType

from contract.models import (
    Contract,
    ContractContributionPlanDetails,
    ContractDetails,
    ContractDetailsMutation,
    ContractMutation,
)


class ContractGQLType(DjangoObjectType):

    class Meta:
        model = Contract
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            "code": ["exact", "istartswith", "icontains", "iexact"],
            **prefix_filterset(
                "policy_holder__", PolicyHolderGQLType._meta.filter_fields
            ),
            "amount_notified": ["exact", "lt", "lte", "gt", "gte"],
            "amount_rectified": ["exact", "lt", "lte", "gt", "gte"],
            "amount_due": ["exact", "lt", "lte", "gt", "gte"],
            "date_payment_due": ["exact", "lt", "lte", "gt", "gte"],
            "state": ["exact"],
            "payment_reference": ["exact", "istartswith", "icontains", "iexact"],
            "amendment": ["exact"],
            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return Contract.get_queryset(queryset, info)

    amount = graphene.Float()


class ContractDetailsGQLType(DjangoObjectType):

    class Meta:
        model = ContractDetails
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            **prefix_filterset("contract__", ContractGQLType._meta.filter_fields),
            **prefix_filterset("insuree__", InsureeGQLType._meta.filter_fields),
            **prefix_filterset(
                "contribution_plan_bundle__",
                ContributionPlanBundleGQLType._meta.filter_fields,
            ),
            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(cls, queryset, info):
            return ContractDetails.get_queryset(queryset, info)


class ContractContributionPlanDetailsGQLType(DjangoObjectType):

    class Meta:
        model = ContractContributionPlanDetails
        interfaces = (graphene.relay.Node,)
        filter_fields = {
            "id": ["exact"],
            **prefix_filterset(
                "contract_details__", ContractDetailsGQLType._meta.filter_fields
            ),
            **prefix_filterset(
                "contribution_plan__", ContributionPlanGQLType._meta.filter_fields
            ),
            **prefix_filterset("contribution__", PremiumGQLType._meta.filter_fields),
            "date_created": ["exact", "lt", "lte", "gt", "gte"],
            "date_updated": ["exact", "lt", "lte", "gt", "gte"],
            "is_deleted": ["exact"],
            "version": ["exact"],
        }

        connection_class = ExtendedConnection

        @classmethod
        def get_queryset(clscls, queryset, info):
            return ContractContributionPlanDetails.get_queryset(queryset, info)


class ContractMutationGQLType(DjangoObjectType):
    class Meta:
        model = ContractMutation


class ContractDetailsMutationGQLType(DjangoObjectType):
    class Meta:
        model = ContractDetailsMutation



from contribution.models import Premium
from contribution_plan.models import ContributionPlan, ContributionPlanBundle
from core import fields
from core import models as core_models
from django.conf import settings
from django.db import models
from graphql import ResolveInfo
from insuree.models import Insuree
from policy.models import Policy
from policyholder.models import PolicyHolder


class ContractManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractManager, self).filter(*args, **kwargs)


class Contract(core_models.HistoryBusinessModel):
    code = models.CharField(db_column="Code", max_length=64, null=False)
    policy_holder = models.ForeignKey(
        PolicyHolder,
        db_column="PolicyHolderUUID",
        on_delete=models.deletion.DO_NOTHING,
        blank=True,
        null=True,
    )
    amount_notified = models.FloatField(
        db_column="AmountNotified", blank=True, null=True
    )
    amount_rectified = models.FloatField(
        db_column="AmountRectified", blank=True, null=True
    )
    amount_due = models.FloatField(db_column="AmountDue", blank=True, null=True)
    date_approved = fields.DateTimeField(
        db_column="DateApproved", blank=True, null=True
    )
    date_payment_due = fields.DateField(
        db_column="DatePaymentDue", blank=True, null=True
    )
    state = models.SmallIntegerField(db_column="State", blank=True, null=True)
    payment_reference = models.CharField(
        db_column="PaymentReference", max_length=255, blank=True, null=True
    )
    amendment = models.IntegerField(
        db_column="Amendment", blank=False, null=False, default=0
    )

    objects = ContractManager()

    @property
    def amount(self):
        amount = 0
        if self.state in [1, 2]:
            amount = self.amount_notified
        elif self.state in [4, 11, 3]:
            amount = self.amount_rectified
        elif self.state in [5, 6, 7, 8, 9, 10]:
            amount = self.amount_due
        else:
            amount = self.amount_due
        return amount

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = "tblContract"

    STATE_REQUEST_FOR_INFORMATION = 1
    STATE_DRAFT = 2
    STATE_OFFER = 3
    STATE_NEGOTIABLE = 4
    STATE_EXECUTABLE = 5
    STATE_ADDENDUM = 6
    STATE_EFFECTIVE = 7
    STATE_EXECUTED = 8
    STATE_DISPUTED = 9
    STATE_TERMINATED = 10
    STATE_COUNTER = 11

    def contract_business_validity(self, prefix=''):
        return [
            models.Q(**{f'{prefix}date_valid_to__isnull': True})
            | models.Q(**{f'{prefix}date_valid_to__gte': self.date_valid_to}),
            models.Q(**{f'{prefix}date_valid_from__lte': self.date_valid_to}),
        ]


class ContractDetailsManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractDetailsManager, self).filter(*args, **kwargs)


class ContractDetails(core_models.HistoryModel):
    contract = models.ForeignKey(
        Contract, db_column="ContractUUID", on_delete=models.deletion.CASCADE
    )
    insuree = models.ForeignKey(
        Insuree, db_column="InsureeID", on_delete=models.deletion.DO_NOTHING
    )
    contribution_plan_bundle = models.ForeignKey(
        ContributionPlanBundle,
        db_column="ContributionPlanBundleUUID",
        on_delete=models.deletion.DO_NOTHING,
    )

    json_param = models.JSONField(db_column="Json_param", blank=True, null=True)

    objects = ContractDetailsManager()

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = "tblContractDetails"


class ContractContributionPlanDetailsManager(models.Manager):
    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ContractContributionPlanDetailsManager, self).filter(
            *args, **kwargs
        )


class ContractContributionPlanDetails(core_models.HistoryBusinessModel):
    contribution_plan = models.ForeignKey(
        ContributionPlan,
        db_column="ContributionPlanUUID",
        on_delete=models.deletion.DO_NOTHING,
    )
    policy = models.ForeignKey(
        Policy,
        db_column="PolicyID",
        on_delete=models.deletion.DO_NOTHING,
        blank=True,
        null=True,
    )
    contract_details = models.ForeignKey(
        ContractDetails,
        db_column="ContractDetailsUUID",
        on_delete=models.deletion.CASCADE,
    )
    contribution = models.ForeignKey(
        Premium,
        db_column="ContributionId",
        related_name="contract_contribution_plan_details",
        on_delete=models.deletion.DO_NOTHING,
        blank=True,
        null=True,
    )
    amount = models.DecimalField(
        db_column='Amount',
        max_digits=18, decimal_places=2, blank=True, null=True)

    objects = ContractContributionPlanDetailsManager()

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            pass
        return queryset

    class Meta:
        db_table = "tblContractContributionPlanDetails"


class ContractMutation(core_models.UUIDModel, core_models.ObjectMutation):
    contract = models.ForeignKey(Contract, models.DO_NOTHING, related_name="mutations")
    mutation = models.ForeignKey(
        core_models.MutationLog, models.DO_NOTHING, related_name="contracts"
    )

    class Meta:
        managed = True
        db_table = "contract_contractMutation"


class ContractDetailsMutation(core_models.UUIDModel, core_models.ObjectMutation):
    contract_detail = models.ForeignKey(
        ContractDetails, models.DO_NOTHING, related_name="mutations"
    )
    mutation = models.ForeignKey(
        core_models.MutationLog, models.DO_NOTHING, related_name="contract_details"
    )

    class Meta:
        managed = True
        db_table = "contract_contractDetailsMutation"



# Generated by Django 3.0.3 on 2021-01-18 14:27

import core.fields
import datetime
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0008_auto_20201230_1052'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contract',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_valid_from',
            field=core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_valid_to',
            field=core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True),
        ),
        migrations.AlterField(
            model_name='contractcontributionplandetails',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='contractcontributionplandetails',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
        migrations.AlterField(
            model_name='contractdetails',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='contractdetails',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_valid_from',
            field=core.fields.DateTimeField(db_column='DateValidFrom', default=datetime.datetime.now),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_valid_to',
            field=core.fields.DateTimeField(blank=True, db_column='DateValidTo', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractcontributionplandetails',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractcontributionplandetails',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractdetails',
            name='date_created',
            field=core.fields.DateTimeField(db_column='DateCreated', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontractdetails',
            name='date_updated',
            field=core.fields.DateTimeField(db_column='DateUpdated', null=True),
        ),
    ]



# Generated by Django 3.0.3 on 2020-12-21 15:47

import core.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contract', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='contract',
            name='amendment',
            field=models.IntegerField(blank=True, db_column='Amendment', null=True),
        ),
        migrations.AddField(
            model_name='contract',
            name='code',
            field=models.CharField(db_column='Code', default='CODE1', max_length=255),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='historicalcontract',
            name='amendment',
            field=models.IntegerField(blank=True, db_column='Amendment', null=True),
        ),
        migrations.AddField(
            model_name='historicalcontract',
            name='code',
            field=models.CharField(db_column='Code', default='CODE2', max_length=255),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='contract',
            name='amount_due',
            field=models.FloatField(blank=True, db_column='AmountDue', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='amount_notified',
            field=models.FloatField(blank=True, db_column='AmountNotified', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='amount_rectified',
            field=models.FloatField(blank=True, db_column='AmountRectified', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_approved',
            field=core.fields.DateTimeField(blank=True, db_column='DateApproved', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='date_payment_due',
            field=core.fields.DateField(blank=True, db_column='DatePaymentDue', null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='contract',
            name='state',
            field=models.SmallIntegerField(blank=True, db_column='State', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='amount_due',
            field=models.FloatField(blank=True, db_column='AmountDue', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='amount_notified',
            field=models.FloatField(blank=True, db_column='AmountNotified', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='amount_rectified',
            field=models.FloatField(blank=True, db_column='AmountRectified', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_approved',
            field=core.fields.DateTimeField(blank=True, db_column='DateApproved', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='date_payment_due',
            field=core.fields.DateField(blank=True, db_column='DatePaymentDue', null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='payment_reference',
            field=models.CharField(blank=True, db_column='PaymentReference', max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='historicalcontract',
            name='state',
            field=models.SmallIntegerField(blank=True, db_column='State', null=True),
        ),
    ]



import datetime

from contribution.models import Premium
from contribution_plan.tests.helpers import (
    create_test_contribution_plan,
    create_test_contribution_plan_bundle,
)
from core.models import User
from insuree.test_helpers import create_test_insuree
from policy.test_helpers import create_test_policy
from policyholder.tests.helpers import create_test_policy_holder
from product.test_helpers import create_test_product

from contract.models import Contract, ContractContributionPlanDetails, ContractDetails


def create_test_contract(policy_holder=None, custom_props={}):
    if not policy_holder:
        policy_holder = create_test_policy_holder()

    user = __get_or_create_simple_contract_user()

    object_data = {
        "code": "CON",
        "policy_holder": policy_holder,
        "amount_notified": 0,
        "amount_rectified": 0,
        "amount_due": 0,
        "date_payment_due": datetime.date(2011, 10, 31),
        "state": 1,
        "payment_reference": "Payment Reference",
        "json_ext": {},
        **custom_props,
    }

    contract = Contract(**object_data)
    contract.save(username=user.username)

    return contract


def create_test_contract_details(
    contract=None, insuree=None, contribution_plan_bundle=None, custom_props={}
):
    if not contract:
        contract = create_test_contract()

    if not insuree:
        insuree = create_test_insuree()

    if not contribution_plan_bundle:
        contribution_plan_bundle = create_test_contribution_plan_bundle()

    user = __get_or_create_simple_contract_user()
    object_data = {
        "contract": contract,
        "insuree": insuree,
        "contribution_plan_bundle": contribution_plan_bundle,
        "json_param": {},
        **custom_props,
    }

    contract_details = ContractDetails(**object_data)
    contract_details.save(username=user.username)

    return contract_details


def create_test_contract_contribution_plan_details(
    contribution_plan=None,
    policy=None,
    contract_details=None,
    contribution=None,
    custom_props={},
):
    if not contribution_plan:
        contribution_plan = create_test_contribution_plan()

    if not policy:
        policy = create_test_policy(
            product=create_test_product(
                "TestCode",
                custom_props={
                    "insurance_period": 12,
                },
            ),
            insuree=create_test_insuree(),
        )

    if not contract_details:
        contract_details = create_test_contract_details()

    if not contribution:
        contribution = Premium.objects.create(
            **{
                "policy_id": policy.id,
                "payer_id": None,
                "amount": 1000,
                "receipt": "Test receipt",
                "pay_date": "2019-01-01",
                "validity_from": "2019-01-01",
                "audit_user_id": 1,
                "pay_type": "C",
            }
        )

    user = __get_or_create_simple_contract_user()
    object_data = {
        "contribution_plan": contribution_plan,
        "policy": policy,
        "contract_details": contract_details,
        "contribution": contribution,
        "json_ext": {},
        **custom_props,
    }

    contract_contribution_plan_details = ContractContributionPlanDetails(**object_data)
    contract_contribution_plan_details.save(username=user.username)

    return contract_contribution_plan_details


def __get_or_create_simple_contract_user():
    if not User.objects.filter(username="admin").exists():
        User.objects.create_superuser(username="admin", password="S/pe®Pąßw0rd™")
    user = User.objects.filter(username="admin").first()
    return user



from calculation.services import run_calculation_rules
from core.signals import Signal
from django.conf import settings
from django.core.mail import BadHeaderError, send_mail
from django.db.models import Q, Subquery
from django.db.models.signals import post_save
from django.dispatch import receiver
from insuree.apps import InsureeConfig
from insuree.models import InsureePolicy
from insuree.signals import signal_before_insuree_policy_query
from payment.apps import PaymentConfig
from payment.models import Payment, PaymentDetail
from payment.signals import signal_before_payment_query
from policy.signals import signal_check_formal_sector_for_policy
from policyholder.apps import PolicyholderConfig
from policyholder.models import PolicyHolderUser

from .config import get_message_approved_contract
from .models import Contract, ContractContributionPlanDetails

_contract_signal_params = ["contract", "user"]
_contract_approve_signal_params = [
    "contract",
    "user",
    "contract_details_list",
    "service_object",
    "payment_service",
    "ccpd_service",
]
signal_contract = Signal(_contract_signal_params)
signal_contract_approve = Signal(_contract_signal_params)


def on_contract_signal(sender, **kwargs):
    contract = kwargs["contract"]
    user = kwargs["user"]
    __save_or_update_contract(contract=contract, user=user)
    return f"contract updated - state {contract.state}"


def on_contract_approve_signal(sender, instance, **kwargs):
    # approve scenario

    user = kwargs["user"]
    contract_service = kwargs["service_object"]
    payment_service = kwargs["payment_service"]
    ccpd_service = kwargs["ccpd_service"]
    # contract valuation
    contract_contribution_plan_details = contract_service.contract_valuation(
        instance, save=True
    )
    if not contract_contribution_plan_details["success"]:
        return contract_contribution_plan_details
    instance.amount_due = contract_contribution_plan_details["data"][
        "total_amount"
    ]
    contribution_result = ccpd_service.create_contribution(
        contract_contribution_plan_details["data"]
    )
    if not contribution_result.get('success', False):
        return contribution_result
    result_payment = __create_payment(
        instance, payment_service, contract_contribution_plan_details["data"]
    )
    if not result_payment.get('success', False):
        return result_payment
    # STATE_EXECUTABLE
    from core import datetime

    now = datetime.datetime.now()
    instance.date_approved = now
    instance.state = 5
    approved_contract = __save_or_update_contract(
        contract=instance, user=user
    )
    if instance.policy_holder.email:
        email_contact_name = (
            instance.policy_holder.contact_name["contactName"]
            if (
                instance.policy_holder.contact_name
                and "contactName" in instance.policy_holder.contact_name
            )
            else instance.policy_holder.trade_name
            or instance.policy_holder.code
        )
        __send_email_notify_payment(
            code=instance.code,
            name=instance.policy_holder.trade_name,
            contact_name=email_contact_name,
            amount_due=instance.amount_due,
            payment_reference=instance.payment_reference,
            email=instance.policy_holder.email,
        )
    return approved_contract


# additional filters for payment in 'contract' tab
def append_contract_filter(sender, **kwargs):
    user = kwargs.get("user", None)
    additional_filter = kwargs.get("additional_filter", None)
    if "contract" in additional_filter:
        # then check perms
        if user.has_perms(PaymentConfig.gql_query_payments_perms) or user.has_perms(
            PolicyholderConfig.gql_query_payment_portal_perms
        ):
            contract_id = additional_filter["contract"]
            contract_to_process = Contract.objects.filter(id=contract_id).first()

            if not contract_to_process or not contract_to_process.policy_holder:
                return Q(pk=-1)
            # check if user is linked to ph in policy holder user table
            type_user = f"{user}"
            # related to user object output (i) or (t)
            # check if we have interactive user from current context
            if "(i)" in type_user:
                from core import datetime

                now = datetime.datetime.now()

                ph_user = PolicyHolderUser.objects.filter(
                    Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=now),
                    date_valid_from__lte=now,
                    policy_holder__id=contract_to_process.policy_holder.id,
                    user__id=user.id,
                ).first()

                if ph_user or user.has_perms(PaymentConfig.gql_query_payments_perms):
                    return Q(
                        **{
                            'payment_details__'
                            + 'premium__'
                            + 'contract_contribution_plan_details__'
                            + 'contract_details__contract__id': contract_id
                        }
                    )


# additional filters for InsureePolicy in contract
def append_contract_policy_insuree_filter(sender, **kwargs):
    user = kwargs.get("user", None)
    additional_filter = kwargs.get("additional_filter", None)
    if "contract" in additional_filter:
        # then check perms
        if user.has_perms(
            InsureeConfig.gql_query_insuree_policy_perms
        ) or user.has_perms(PolicyholderConfig.gql_query_insuree_policy_portal_perms):
            contract_id = additional_filter["contract"]
            contract_to_process = Contract.objects.filter(id=contract_id).first()
            if not contract_to_process or not contract_to_process.policy_holder:
                return Q(pk=-1)
            # check if user is linked to ph in policy holder user table
            type_user = f"{user}"
            # related to user object output (i) or (t)
            # check if we have interactive user from current context
            if "(i)" in type_user:
                from core import datetime

                now = datetime.datetime.now()
                ph_user = PolicyHolderUser.objects.filter(
                    Q(date_valid_to__isnull=True) | Q(date_valid_to__gte=now),
                    date_valid_from__lte=now,
                    policy_holder__id=contract_to_process.policy_holder.id,
                    user__id=user.id,
                ).first()

                if ph_user or user.has_perms(
                    InsureeConfig.gql_query_insuree_policy_perms
                ):
                    policies = list(
                        ContractContributionPlanDetails.objects.filter(
                            contract_details__contract__id=contract_id
                        ).values_list("policy", flat=True)
                    )

                    return Q(
                        start_date__gte=contract_to_process.date_valid_from,
                        start_date__lte=contract_to_process.date_valid_to,
                        policy__in=policies,
                    )


# check if policy is related to formal sector contract
def formal_sector_policies(sender, **kwargs):
    policy_id = kwargs.get("policy_id", None)
    ccpd = ContractContributionPlanDetails.objects.filter(
        policy__id=policy_id, is_deleted=False
    ).first()
    if ccpd:
        cd = ccpd.contract_details
        contract = cd.contract
        return contract.policy_holder
    else:
        return None


signal_contract.connect(on_contract_signal, dispatch_uid="on_contract_signal")
signal_contract_approve.connect(
    on_contract_approve_signal, dispatch_uid="on_contract_approve_signal"
)
signal_before_payment_query.connect(append_contract_filter)
signal_before_insuree_policy_query.connect(append_contract_policy_insuree_filter)
signal_check_formal_sector_for_policy.connect(formal_sector_policies)


@receiver(post_save, sender=Payment, dispatch_uid="payment_signal_paid")
def activate_contracted_policies(sender, instance, **kwargs):
    received_amount = instance.received_amount if instance.received_amount else 0
    # check if payment is related to the contract
    payment_detail = (
        PaymentDetail.objects.filter(payment__id=int(instance.id))
        .prefetch_related(
            "premium__contract_contribution_plan_details__contract_details__contract"
        )
        .prefetch_related("premium__contract_contribution_plan_details")
        .filter(premium__contract_contribution_plan_details__isnull=False)
    )
    if len(list(payment_detail)) > 0:
        if instance.expected_amount <= received_amount:
            contribution_list_id = list(set(pd.premium.id for pd in payment_detail))
            contract_list = Contract.objects.filter(
                contractdetails__contractcontributionplandetails__contribution__id__in=contribution_list_id
            ).distinct()
            ccpd_full_list = ContractContributionPlanDetails.objects.filter(
                contract_details__contract__in=Subquery(contract_list.values('id'))
            ).prefetch_related('contract_details')

            # 1- check if the contract have payment attached to each contributions
            # (nbr CCPD of all contract in step 0= Paymentdetails in Steps 0)
            if len(ccpd_full_list) == len(list(payment_detail)):
                for contract in contract_list:
                    if contract.state == Contract.STATE_EXECUTABLE:
                        # get the ccpd related to the currenttly processing contract
                        ccpd_list = [
                            ccpd for ccpd in ccpd_full_list if ccpd.contract_details.contract_id == contract.id
                        ]
                        # TODO support Splitted payment and check that
                        #  the payment match the value of all contributions
                        for ccpd in ccpd_list:
                            members = run_calculation_rules(
                                ccpd, "members", contract.user_updated
                            )
                            for insuree in members:
                                InsureePolicy.objects.create(
                                    **{
                                        "insuree": insuree,
                                        "policy": ccpd.policy,
                                        "enrollment_date": ccpd.date_valid_from,
                                        "start_date": ccpd.date_valid_from,
                                        "effective_date": ccpd.date_valid_from,
                                        "expiry_date": ccpd.date_valid_to,
                                        "audit_user_id": -1,
                                    }
                                )
                        contract.state = Contract.STATE_EFFECTIVE
                        __save_or_update_contract(contract, contract.user_updated)


def __save_json_external(user_id, datetime, message):
    return {
        "comments": [
            {"From": "Portal/webapp", "user": user_id, "date": datetime, "msg": message}
        ]
    }


def __save_or_update_contract(contract, user):
    contract.save(username=user.username)
    historical_record = contract.history.all().first()
    contract.json_ext = __save_json_external(
        user_id=str(historical_record.user_updated.id),
        datetime=str(historical_record.date_updated),
        message=f"contract updated - state " f"{historical_record.state}",
    )
    contract.save(username=user.username)
    return contract


def __create_payment(contract, payment_service, contract_cpd):
    from core import datetime

    now = datetime.datetime.now()
    # format payment data
    payment_data = {
        "expected_amount": contract.amount_due,
        "request_date": now,
    }
    payment_details_data = payment_service.collect_payment_details(
        contract_cpd["contribution_plan_details"]
    )
    return payment_service.create(
        payment=payment_data, payment_details=payment_details_data
    )


def __send_email_notify_payment(
    code, name, contact_name, amount_due, payment_reference, email
):
    try:
        email = send_mail(
            subject="Contract payment notification",
            message=get_message_approved_contract(
                language=settings.LANGUAGE_CODE.split("-")[0],
                code=code,
                name=name,
                contact_name=contact_name,
                due_amount=amount_due,
                payment_reference=payment_reference,
            ),
            from_email=settings.EMAIL_HOST_USER,
            recipient_list=[email],
            fail_silently=False,
        )
        return email
    except BadHeaderError:
        return ValueError("Invalid header found.")
